; Some common reporting code     -*- mode: scheme; mode: font-lock -*-
; 2000-07-23 / lth

(require 'list)

; `Results' is the output of basic-statistics: a list of assoc lists.
; `Extractors' is a list of of functions that take a unary GET and
;    a benchmark designator and return a value.
;
; Returns a list of lists of data, each sublist has the same benchmark
; designator in every datum.  Each datum is a list of length at least 2:
; benchmark designator, system info, and data values.  If not all the
; iterations succeeded, then there are no data values.

(define (extract-data results . extractors)

  (define (get x datum)
    (let ((probe (assq x datum)))
      (if probe
          (cdr probe)
          (error "get: Can't find " x " in " datum))))

  (define (group-by-benchmark-and-setting-type results)
    (accumulate comparable-runs? results))

  (map (lambda (group-of-results)
         (map (lambda (result)
                (let ((bm  (benchmark-designator result))
                      (get (lambda (tag) 
                             (get tag result))))
                  (if (all-iterations-succeeded? result)
                      `(,bm
                        ,(get 'system-info)
                        ,@(map (lambda (f) (f get bm)) extractors))
                      `(,bm 
                        ,(get 'system-info)))))
              group-of-results))
       (group-by-benchmark-and-setting-type results)))


; `Prefix' is a file name prefix.
; `Get-x' and `get-y' are unary functions that return X and Y values when
;    applied to a datum.
; `Data' has the following syntax:
;
;   data ::= ((datum ...) ...)
;   datum ::= (bm . some data)
;
; Each sublist of `data' should have data with the same value of BM, and 
; no two sublists of `data' should have the same value of BM.

(define (plot-benchmarks prefix data get-x get-y)

  ; Relies on x.0  being printed with the zero (Gambit doesn't do that
  ; but Chez and Larceny do).

  (define (twoplaces n)
    (let ((n (round (* n 100.0))))
      (if (integer? (/ n 10.0))
          (string-append (number->string (/ n 100.0)) "0")
          (number->string (/ n 100.0)))))

  (let ((previous #f)
        (out      #f)
        (data     (apply append data)))
    (for-each
     (lambda (x)
       (let ((bm (car x)))
         (if (not (equal? bm previous))
             (begin (if previous
                        (close-output-port out))
                    (set! out (open-output-file 
                               (string-append 
                                prefix
                                (format #f "~a~a"
                                        (car bm)
                                        (sanitize-string
                                         (symbol->string (cadr bm)))))))
                    (set! previous bm)))
         (format out "~a ~a~%" 
                 (twoplaces (get-x x))
                 (twoplaces (get-y x)))))
     data)
    (if out (close-output-port out))))


; REPORT is a function that takes the output of `basic-statistics' and
; returns a value as from `extract-data', above.  (That is, a list of
; lists of data, each sublist has the same benchmark designator in every
; datum.  Each datum is a list of length at least 2: benchmark
; designator, system info, and data values.  If not all the iterations
; succeeded, then there are no data values.)
;
; RAW-RESULTS is a list of raw data sets.  (Anachronism, sorry.)
;
; COMPUTE is a function that is applied to each nonempty list of data 
; values returned from REPORT, less benchmark name and system info.
; It should return a single value.
;
; Returns a list of lists of pairs.  Each sublist consists of pairs of
; benchmark designator and value returned from COMPUTE; every element in
; a sublist has the same benchmark designator.

(define (compute-and-accumulate-statistics report raw-results compute)
  (accumulate-unordered
   (lambda (a b) (equal? (car a) (car b)))
   (filter-map
    (lambda (x)
      (if (not (null? (cddr x)))
          (cons (car x) (apply compute (cddr x)))
          #f))
    (mappend 
     (lambda (r)
       (apply append (report (basic-statistics r))))
     raw-results))))


; Exactly as above, but DATA is instead a standard list of canonical
; data, as produced by 2gen-canonical-data.

(define (new-compute-and-accumulate-statistics report data compute)
  (accumulate-unordered
   (lambda (a b) (equal? (car a) (car b)))
   (filter-map
    (lambda (x)
      (if (not (null? (cddr x)))
          (cons (car x) (apply compute (cddr x)))
          #f))
    (mappend 
     (lambda (r)
       (apply append (report (car r))))
     data))))


; Generic mark/cons reporting.  REPORT is a reporting function acceptable
; to compute-and-accumulate-statistics.

(define (create-vs-mark/cons-report fn report raw-results)

  ; Relies on x.0  being printed with the zero (Gambit doesn't do that
  ; but Chez and Larceny do).

  (define (twoplaces n)
    (let ((n (round (* n 100.0))))
      (if (integer? (/ n 10.0))
          (string-append (number->string (/ n 100.0)) "0")
          (number->string (/ n 100.0)))))

  (call-with-output-file fn
    (lambda (out)
      (format out "\\begin{tabular}{|l|rrrrrr|}~%")
      (format out "\\hline~%")
      (format 
       out 
       " Program & $n$ & min & max & mean & median & std.dev. \\\\~%")
      (format out "\\hline~%")
      (for-each
       (lambda (x)
         (format out " ~a~a & ~a & ~a & ~a & ~a & ~a & ~a \\\\~%"
                 (car (caar x)) 
                 (cadr (caar x))
                 (length x)
                 (twoplaces (apply min (map cdr x)))
                 (twoplaces (apply max (map cdr x)))
                 (twoplaces (mean (map cdr x)))
                 (twoplaces (median (map cdr x)))
                 (twoplaces (standard-deviation (map cdr x)))))
       (compute-and-accumulate-statistics report raw-results /))
      (format out "\\hline~%")
      (format out "\\end{tabular}~%"))))


(define (create-benchmark-2d-plot filename-prefix report raw-results)
  (let ((get-x cadr)
        (get-y cddr))
    (plot-benchmarks 
     filename-prefix
     (map (lambda (data) 
            (sort-by-x-and-y data get-x get-y))
          (compute-and-accumulate-statistics report raw-results cons))
     get-x
     get-y)))

(define (new-create-benchmark-2d-plot filename-prefix report data)
  (let ((get-x cadr)
        (get-y cddr))
    (plot-benchmarks 
     filename-prefix
     (map (lambda (data) 
            (sort-by-x-and-y data get-x get-y))
          (new-compute-and-accumulate-statistics report data cons))
     get-x
     get-y)))


; Results as reported by basic-statistics.

(define (report-copy-speed trend-data) ; kilobytes/millisecond.
  (let ((results (car trend-data))
        (n       (cadr trend-data)))
    (filter-map 
     (lambda (r)
       (let ((get (lambda (tag)
                    (let ((probe (assq tag r)))
                      (if probe
                          (cdr probe)
                          (error "report-copy-speed: can't find " tag
                                 " in " r))))))
         (if (all-iterations-succeeded? r)
             (if (= n 5)
                 (round (/ (round (/ (get 'run-bytes-copied-best) 1024.0))
                           (get 'run-cpu-gc-best)))
                 (round (/ (round (/ (get 'bytes-copied-best) 1024.0))
                           (get 'cpu-gc-best))))
             #f)))
     results)))


; Histogram for data where the length of the list is the magnitude and
; the car of a list is a representative element.  No empty lists allowed.

(define (histogram width data)          ; data ::= ((x x ...) ...)
  (let ((highest (apply max (map length data))))
    (for-each (lambda (d)
                (display (tf (format #f "~a" (car d))))
                (display " ")
                (display (make-string 
                          (inexact->exact 
                           (round (* width (/ (length d) highest))))
                          #\+))
                (newline))
              data)))

; eof
