; Comparative trend reporting        -*- mode: scheme; mode: font-lock -*-
; 2000-11-19 / lth

(require 'control)
(require 'list)
(require 'number)
(require 'format)

; Computes A as a percentage of B.  This function presents a skewed
; visual image because the range of the x axis is [-100.0,+inf.0), and
; if A is 10 times slower than B, then it is 10 units to the right of
; the origin, whereas if A is 10 times faster than B, then it is 0.9 units 
; to the left of the origin.

'(define comparing-function
  (lambda (a b)
    (* 100 (- (/ a b) 1.0))))

; Computes B as a ratio of A along the negative x axis, and A as a ratio
; of B along the positive x axis.  The interpretation is again a percentage,
; but it is the larger as a percentage of the smaller, and the percentage
; is always positive.  The range of the x azis is (-inf.0,+inf.0).

(define comparing-function
  (lambda (a b)
    (if (< a b)
        (* 100 (- 1.0 (/ b a)))
        (* 100 (- (/ a b) 1.0)))))

; Warning: Complicated.  Sorry.
;
; The idea here is that if we're given two data sets, each representing
; a distinct collector class but each containing several sets of
; measurements for the collector class, we're going to be doing a 'best
; of class' comparison.  
;
; For example, we can compare all the GEN collectors against all the DOF
; collectors by comparing the data set consisting of measurements from
; 2GEN and 3GEN against that consisting of measurements from 1DOF thru
; 4DOF, like this:
;
;  (comparative-trend (list (basic-statistics 2gen-upper-1)
;                           (basic-statistics 3gen-upper-5))
;                     (list (dof-compute-canonical-data 1dof-upper-1)
;                           (dof-compute-canonical-data 2dof-upper-5)
;                           (dof-compute-canonical-data 3dof-upper-5)
;                           (dof-compute-canonical-data 4dof-upper-5))
;                     (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
;                     (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
;                     min)
;
; Given: two data sets GROUP1 and GROUP2, which are lists of (data as
; produced by basic-statistics), that is, lists of lists of association
; lists; and two value extractors VALUE-EXTRACTOR1 and VALUE-EXTRACTOR2
; (one for each data set); and a function BEST-OF-N that takes a
; nonempty list of values as returned by the extractors and returns the
; best value in the set:
;
; With the following hairy constraints: the sublists of GROUP1 and
; GROUP2 must be pairwise of the same length; and the runs of each
; sublist must be pairwise comparable; and the sublists of GROUP1
; (GROUP2) must all be the same length; and the runs of each pair of
; sublists of GROUP1 (GROUP2) must be pairwise comparable:
;
; Then: computes the best result for each benchmark at a particular
; setting across the results in a group, and compares the results from
; the two groups using `comparing-function', above, resulting in a list
; of advantage numbers for the systems (negative means group1, positive
; means group2), and then sorts these numbers into bins using 
; `bin-number', below, and finally returns a list of lists sorted by the 
; bin-number of each sublist.  The advantage data are not reduced to
; bin data.
;
; Of course, it works even if only one set of runs is in each group,
; and that is probably where its real utility lies.
;
; Note: The two extractors is an anachronism; one would be sufficient.

(define (comparative-trend group1 group2 value-extractor1 value-extractor2 
                           best-of-n)

  (define (compute stats extractor)
    (map (lambda (item)
           (if (all-iterations-succeeded? item)
               (extractor item)
               #f))
         stats))

  (define (best xs)
    (letrec ((the-best 
              (lambda (xs)
                (if (null? xs)
                    #f
                    (apply best-of-n xs)))))
      (if (null? (car xs))
          '()
          (cons (the-best (filter-map car xs)) (best (map cdr xs))))))

  (define (compute-statistics)
    (filter-map
     (lambda (a b) 
       (if (and a b)
           (comparing-function a b)
           #f))
     (best (map (lambda (stats) (compute stats value-extractor1)) 
                group1))
     (best (map (lambda (stats) (compute stats value-extractor2)) 
                group2))))

  ; If comparative-trends returns garbage, all is lost.  So be paranoid.
  (define (sanity-check)
    (if (not (call-with-current-continuation
              (lambda (k)
                (for-each 
                 (lambda (d1)
                   (for-each
                    (lambda (d2)
                      (if (not (and (= (length d1) (length d2))
                                    (every? comparable-runs? d1 d2)))
                          (k #f)))
                    group2))
                 group1)
                #t)))
        (begin
          (set! err1 group1)
          (set! err2 group2)
          (error "Precondition for comparative-trend violated.  "
                 "Data are in ERR1 and ERR2."))))

  (sanity-check)
  (accumulate
   (lambda (a b) (= (bin-number a) (bin-number b)))
   (sort
    (compute-statistics)
    (lambda (a b) (< (bin-number a) (bin-number b))))))

(define (bin-number percentage)
  (* (sign percentage) 
     (quotient (+ (inexact->exact (truncate (abs percentage))) 1) 2)))

(define (percentage bin-number)
  (* bin-number 2))


; Takes some data sets for a single or multiple collectors and returns
; data for a histogram of pairwise advantages for pairs of runs of
; comparable settings.  The returned value is a list of pairs where
; each pair is an advantage number (a bin) and the number of pairs of
; runs that fell into that bin.
;
; For example:
;  (compare-across-all (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
; 		       (lambda (x) (cdr (assq 'run-cpu-gc-best x)))
; 	 	       (all-2gen-canonical-data)
;		       (all-3dof-canonical-data)
;		       min)
;
; EXTRACT1 and EXTRACT5 are extractors for list of 1 iterations and 5
; iterations respectively (an anachronism).
;
; AS and BS are data sets as returned by `ALL-*-CANONICAL-DATA', see
; comparative-reports.sch.
;
; BEST-OF-N is a procedure that takes a nonempty list of values as
; returned by the extractors and returns the best of them.

(define (compare-across-all extract1 extract5 as bs best-of-n)
  (map (lambda (x) 
         (cons (percentage (bin-number (car x))) (length x)))
       (merge
        (lambda (a b) 
          (< (bin-number (car a)) (bin-number (car b))))
        append
        (do ((i  0 (+ i 1))
             (xa (car as) (cdr xa))
             (xb (car bs) (cdr xb))
             (l '() (cons (comparative-trend
                           (map (lambda (a) (car (list-ref a i))) as)
                           (map (lambda (b) (car (list-ref b i))) bs)
                           (if (= 1 (cadar xa)) extract1 extract5)
                           (if (= 1 (cadar xb)) extract1 extract5)
                           best-of-n)
                          l)))
            ((null? xa) (reverse l))))))


; This is as compare-across-all, but the results are broken down by
; benchmark.  This procedure returns an association list keyed on
; the benchmark designator, where the value is a list of pairs as
; described for compare-across-all.

(define (compare-by-benchmark extract1 extract5 as bs best-of-n)

  ; xsl ::= (xs ...)
  ; xs  ::= (x ...)
  ; x   ::= ((<alist> ...) n)

  (define (select-benchmark xsl bm)
    (let ((iter (car bm))
          (name (cadr bm)))
      (filter-map
       (lambda (xs)
         (let ((nxs
                (filter-map
                 (lambda (x)
                   (if (not (= iter (cadr x)))
                       #f
                       (let ((nalists 
                              (filter (lambda (alist)
                                        (eq? name 
                                             (cdr 
                                              (assq 'benchmark-name alist))))
                                      (car x))))
                         (if (not (null? nalists))
                             (list nalists iter)
                             #f))))
                 xs)))
           (if (not (null? nxs))
               nxs
               #f)))
          xsl)))

  (map (lambda (bm)
         (cons bm
               (compare-across-all extract1
                                   extract5 
                                   (select-benchmark as bm)
                                   (select-benchmark bs bm)
                                   best-of-n)))
       benchmark-names))

; Given results produced by compare-by-benchmark, return a list 
; for each benchmark: 
;
;  benchmark name as a string (_not_ designator)
;  minimum value
;  maximum value
;  25th percentile value
;  75th percentile value
;  mean - stdandard deviation
;  mean + standard deviation
;  mean

(define (extract-benchmark-trend-data results)

  (define (expand data)
    (if (null? data)
        '()
        (let ((v (caar data))
              (n (cdar data)))
          (append (make-list n v)
                  (expand (cdr data))))))

  (define (middle-fifty data)
    (let ((total (sum cdr data)))
      (let loop ((sum 0) (data data))
        (if (< (+ sum (cdar data)) (* 0.25 total))
            (loop (+ sum (cdar data)) (cdr data))
            (let ((start (caar data)))
              (let loop ((sum sum) (data data) (end start))
                (if (> sum (* 0.75 total))
                    (values start end)
                    (loop (+ sum (cdar data)) (cdr data) (caar data)))))))))

  (map (lambda (bm)
         (let* ((name     (car bm))
                (data     (cdr bm))
                (smallest (apply min (map car data)))
                (largest  (apply max (map car data)))
                (avg      (mean (expand data))) ; weighted mean
                (dev      (standard-deviation (expand data))))
           (let-values (((75low 75high) (middle-fifty data)))
             (list (format-benchmark-name name)
                   smallest 
                   largest
                   75low
                   75high
                   (- avg dev)
                   (+ avg dev)
                   avg))))
       results))


; DATA1 and DATA2 are single data sets with number-of-iteration
; metadata, as produced by `sc-canonical-data'.
;
; Returns a list: for each pair of comparable runs where both runs
; succeeded, a datum (a . (b . c)) where a is the benchmark identifier,
; b is the CPU time advantage number, and c is the space advantage
; number.  The results are not sorted by anything.

(define (extract-by-benchmark-space-time data1 data2)

  (define (get-cpu-time get bm)
    (if (= 1 (car bm))
        (/ (get 'cpu-best) 1000.0)
        (/ (get 'run-cpu-best) 1000.0)))

  (define (get-heap-size get bm)
    (get 'bytes-max-heap))

  (define (extract-from-data-set set)
    (mappend 
     (lambda (x) 
       (apply append
              (extract-data (car x) get-cpu-time get-heap-size)))
     set))

  (filter-map 
   (lambda (a b)
     (if (and (not (null? (cddr a)))
              (not (null? (cddr b))))
         (cons (car a)
               (cons (comparing-function (caddr a) (caddr b))
                     (comparing-function (cadddr a) (cadddr b))))
         #f))
   (extract-from-data-set data1)
   (extract-from-data-set data2)))

; (As far as I can tell, the following description is almost completely
; detached from reality.  This actually computes advantage numbers, and
; a GET function is passed to the extractor functions.)
;
; DATA1 and DATA2 are single data sets with number-of-iteration
; metadata, as produced by `sc-canonical-data'.  The optional arguments
; are all data extractors: they take a single datum as produced by 
; basic-statistics (an assoc list) and a benchmark designator, and
; return a value.
;
; Returns a list: for each pair of comparable runs where both runs
; succeeded, a datum (a . (n1 ...)) where a is the benchmark identifier
; and the n are numbers returned by the extractors.  The results are
; not sorted by anything.

(define (extract-by-benchmark data1 data2 . extractors)

  (define (extract-from-data-set set)
    (mappend 
     (lambda (x) 
       (apply append (apply extract-data (car x) extractors)))
     set))

  (filter-map 
   (lambda (a b)
     (if (and (not (null? (cddr a)))
              (not (null? (cddr b))))
         (cons (car a)
               (map comparing-function (cddr a) (cddr b)))
         #f))
   (extract-from-data-set data1)
   (extract-from-data-set data2)))


; Qualitative tests

; Experimental, but probably close to right.
; The constant 1.5 should be more sophisticated.

(define (qualitative-test d1 d2)
  (let ((d (extract-by-benchmark-space-time d1 d2))
        (class (vector 0 0 0 0 0 0)))
    (define factor 1.5)
    (define incd #f)
    (define (inc n)
      (set! incd #t)
      (vector-set! class n (+ (vector-ref class n) 1)))
    (for-each (lambda (d)
                (let ((time (cadr d))
                      (space (cddr d)))
                  (let ((time_A_B  (- time))
                        (space_A_B (- space))
                        (time_B_A  time)
                        (space_B_A space))
                    (inc 5)
                    (let ((r1 (and (>= space_A_B (* factor time_B_A))
                                   (>= space_A_B 2.0)))
                          (r2 (and (>= time_A_B (* factor space_B_A))
                                   (>= time_A_B 2.0)))
                          (r3 (and (>= space_B_A (* factor time_A_B))
                                   (>= space_B_A 2.0)))
                          (r4 (and (>= time_B_A (* factor space_A_B))
                                   (>= time_B_A 2.0))))
                      (inc (cond (r1 0)
                                 (r2 1)
                                 (r3 2)
                                 (r4 3)
                                 (else 4)))))))
              d)
    class))

(define (qualitative-test-by-benchmark d1 d2)
  (let ((d (extract-by-benchmark-space-time d1 d2))
        (class (vector 0 0 0 0 0 0)))
    (define factor 1.5)
    (define incd #f)
    (define (inc n)
      (set! incd #t)
      (vector-set! class n (+ (vector-ref class n) 1)))
    (map (lambda (bm)
           (set! class (vector 0 0 0 0 0 0))
           (for-each 
            (lambda (d)
              (let ((time (cadr d))
                    (space (cddr d)))
                (let ((time_A_B  (- time))
                      (space_A_B (- space))
                      (time_B_A  time)
                      (space_B_A space))
                  (inc 5)
                  (let ((r1 (and (>= space_A_B (* factor time_B_A))
                                 (>= space_A_B 2.0)))
                        (r2 (and (>= time_A_B (* factor space_B_A))
                                 (>= time_A_B 2.0)))
                        (r3 (and (>= space_B_A (* factor time_A_B))
                                 (>= space_B_A 2.0)))
                        (r4 (and (>= time_B_A (* factor space_A_B))
                                 (>= time_B_A 2.0))))
                    (inc (cond (r1 0)
                               (r2 1)
                               (r3 2)
                               (r4 3)
                               (else 4)))))))
            (filter (lambda (x) (equal? (car x) bm)) d))
           (list bm class))
         benchmark-names)))


; Reports

(define (report-by-benchmark-space-time file data1 data2) 
  (plot-benchmarks
   file
   (accumulate-unordered
    (lambda (a b) (equal? (car a) (car b)))
    (extract-by-benchmark-space-time data1 data2))
   cadr
   cddr))
         
; For the following, COMPARER is one of compare-by-benchmark, compare-all.
; AS and BS are as returned by all-*-canonical-data.
;
; Returns a value as for the comparer.

(define (compare-by-cpu-time comparer as bs)
  (comparer (lambda (item) (cdr (assq 'cpu-best item)))
            (lambda (item) (cdr (assq 'run-cpu-best item)))
            as
            bs
            min))

(define (compare-by-cpu-gc-time comparer as bs)
  (comparer (lambda (item) (cdr (assq 'cpu-gc-best item)))
            (lambda (item) (cdr (assq 'run-cpu-gc-best item)))
            as
            bs
            min))

(define (compare-by-cpu-mutator-time comparer as bs)
  (comparer (lambda (item) 
              (- (cdr (assq 'cpu-best item)) 
                 (cdr (assq 'cpu-gc-best item))))
            (lambda (item) 
              (- (cdr (assq 'run-cpu-best item)) 
                 (cdr (assq 'run-cpu-gc-best item))))
            as
            bs
            min))
  
; Note, does not include remembered set space.  This is correct.

(define (compare-by-space comparer as bs)
  (comparer (lambda (item) (cdr (assq 'bytes-max-heap item)))
            (lambda (item) (cdr (assq 'bytes-max-heap item)))
            as
            bs
            min))

; Does include remembered set space.

(define (compare-by-total-space comparer as bs)
  (let ((fn (lambda (datum)
              (+ (cdr (assq 'bytes-max-heap datum))
                 (cdr (assq 'bytes-max-remset datum))))))
    (comparer fn fn as bs min)))

(define (compare-all-by-cpu-time as bs)
  (compare-by-cpu-time compare-across-all as bs))

(define (compare-all-by-cpu-gc-time as bs)
  (compare-by-cpu-gc-time compare-across-all as bs))

(define (compare-all-by-cpu-mutator-time as bs)
  (compare-by-cpu-mutator-time compare-across-all as bs))

(define (compare-all-by-space as bs)
  (compare-by-space compare-across-all as bs))

(define (compare-benchmarks-by-cpu-time as bs)
  (compare-by-cpu-time compare-by-benchmark as bs))

(define (compare-benchmarks-by-cpu-gc-time as bs)
  (compare-by-cpu-gc-time compare-by-benchmark as bs))

(define (compare-benchmarks-by-cpu-mutator-time as bs)
  (compare-by-cpu-mutator-time compare-by-benchmark as bs))

(define (compare-benchmarks-by-space as bs)
  (compare-by-space compare-by-benchmark as bs))

(define (compare-benchmarks-by-total-space as bs)
  (compare-by-total-space compare-by-benchmark as bs))


; eof
