; -*- mode: scheme; mode: font-lock -*-
; 2000-07-21 / lth
;
; Parser for stripped output from benchmark runs.
; This version is able to deal with REMSET-PROFILE output also.
;
; Evaluate 
;    (crunch filename) 
; where filename is the name of a file with raw or stripped output from a
; benchmarking run.  This returns a list of "system" data structures; see 
; below.

(require 'list)

; Data structures


; Some data are redundantly recorded because they were recorded by several
; pieces of the benchmarking infrastructure, and some data are not useful
; programatically.

; A _system_ is a set of runs of one benchmark on one Scheme system with
; one set of settings -- essentially, a set of identical invocations of
; a program:

(define (make-system info tags settings benchmark-names runs)
  (list info tags settings benchmark-names runs))

(define (system.info x) (list-ref x 0))
(define (system.tags x) (list-ref x 1))
(define (system.settings x) (list-ref x 2))
(define (system.benchmark-name x) (list-ref x 3))
(define (system.runs x) (list-ref x 4))

(define (system.info! x v) (list-set! x 0 v))

; <info> is a list, it takes two major forms:
;     (<system-tag> variable <load-factor>)
;     (<system-tag> fixed <heap-size> <gc-trigger>)
; where <load-factor> is the load factor (decimal)
;       <heap-size> is the heap size in bytes
;       <gc-trigger> is the gc trigger ratio
;       <system-tag> is a symbol that identifies the system: gen, sc,
;          or a list 
;             (bdw <option> ...) 
;             (dof <n>)
;          For BDW, each <option> identifies an aspect of the collector: 
;               -nip, -aip, -atomic, -noatomic, -byte, -nobyte.
;          Some may be absent; the default is -nip -atomic -nobyte.
;          For DOF, <n> is the requested number of areas.
;
; <tags> is a list of lists: each sublist is 
;     (<tag> <string>) 
; where <tag> is one of identifier, date, host, all of which are 
; human-readable.
;
; <settings> is a list of settings as reported by the scripts: each 
; setting is a list 
;    (<tag> <value>)
; where <tag> is a symbol; one of these forms:
;      (command <string>)             the command line executed
;      (codetype <symbol>)            one of common, bdw, sc, gen
;      (output_file <string>)         file name of output
;      (iterations_per_run <number>)  number of iterations of each program
;      (number_of_runs <number>)      number of runs
;      (benchmarks <list>)            list of benchmark names (symbols)
;
; <benchmark-names> is a list of symbols: the benchmark programs to run
; during this invocation of the script.
;
; <runs> is a list of "run" data structures.  A run is the output of a
; single invocation of a program.  The run data structure does not carry
; all the interesting information about the run, so look at the system
; data in which the run is stored.

(define (make-run run-number command stats ps warnings errors commands 
                  remset-profiles remset-trace)
  (list run-number command stats ps warnings commands errors remset-profiles
        remset-trace))

(define (run.run-number x) (list-ref x 0))
(define (run.command x) (list-ref x 1))
(define (run.stats x) (list-ref x 2))
(define (run.ps x) (list-ref x 3))
(define (run.warnings x) (list-ref x 4))
(define (run.commands x) (list-ref x 5))
(define (run.errors x) (list-ref x 6))
(define (run.remset-profiles x) (list-ref x 7))
(define (run.remset-trace x) 
  (if (> (length x) 8)
      (list-ref x 8)
      '()))

; <run-number> is a list of 2 elements: benchmark name (symbol) and the
; ordinal number of the run within a set of runs.
;
; <command> is a string: the command used to run the program.
;
; <stats> is a list of association lists: one assoc list of measurements
; for each iteration that succeeded.  The assoc lists are those printed
; in the STATS-OUTPUT record plus some additional data computed by this
; program; the keys are
;
;    bytes-allocated
;    bytes-reclaimed
;    bytes-marked
;    elapsed-time
;    cpu-time-user
;    cpu-time-system
;    elapsed-time-gc
;    elapsed-time-major-gc
;    elapsed-time-minor-gc
;    elapsed-time-full-gc
;    cpu-time-gc
;    cpu-time-gc-promotion
;    cpu-time-gc-collection
;    cpu-time-gc-full
;    count-gc
;    count-major-gc
;    count-minor-gc
;    count-full-gc
;    bytes-max-heap-used
;    remset-entries-scanned
;    remset-bytes-scanned
;    remset-ssb-transactions
;    remset-add-transactions
;    has-np-remset                Boolean
;    swb-total-assignments
;    swb-array-assignments
;    swb-filtered-remembered
;    swb-filtered-constant
;    swb-filtered-not-old-to-young
;    swb-transactions
;    raw-data-before              Memstats vector
;    raw-data-after               Memstats vector
;
; Remembered-set data: the data are lists of numbers, one number for
; each remembered set.  If has-np-remset is #t, then the last entry
; in each remset list is data for the nonpredictive "extra" remembered 
; set, which tracks young-to-old pointers.
;
; BDW notes
; * BDW reports elapsed-time-minor-gc and count-minor-gc as 0, as it 
;   does not distinguish major and minor GCs.
; * BDW reports bytes-marked as 0, though this can be fixed.
;
; <ps> is a number: the process size in kilobytes after the last iteration,
; or #f if the run failed during any iteration.
;
; <warnings> is a list of strings: any warning messages printed.
;
; <errors> is a list of strings: any error messages printed.
;
; <commands> is a list of strings: any command lines saved following warnings.
;
; <remset-profiles> is a list of lists of length 2: the number of nanoseconds
; spent in remembered set scanning recorded two ways (sum of costs of scanning
; each set; cost of scanning all sets).  One sublist of numbers for every
; run in run.stats, or an empty list if no such data are available.
;
; <remset-trace> is a list of remembered set trace data.

; Parser

; Each option is a pair (string . proc) where string is any string
; and proc is a procedure that takes a string and returns a string
; or #f.  If the string matches the prefix of an input line then
; proc is called with that line; if it returns a string, then the
; string is processed in place of the original line, and if it returns
; #f then the line is ignored.

(define (crunch filename . options)

  (define v-systems '())
  (define v-begin-system #f)
  (define v-tags '())
  (define v-settings '())
  (define v-benchmark-name #f)
  (define v-runs '())
  (define v-warnings '())
  (define v-errors '())
  (define v-saved-lines '())
  (define v-run-number #f)
  (define v-benchmark-command #f)
  (define v-stats-output '())
  (define v-ps-output #f)
  (define v-remset-profiles '())
  (define v-remset-profile-tmp #f)
  (define v-raw-memstats-tmp #f)
  (define v-remset-trace '())

  (define line-number 0)
  (define errors 0)

  (define (clear-run)
    (set! v-run-number #f)
    (set! v-benchmark-command #f)
    (set! v-stats-output '())
    (set! v-ps-output #f)
    (set! v-remset-profiles '())
    (set! v-remset-profile-tmp #f)
    (set! v-raw-memstats-tmp #f)
    (set! v-warnings '())
    (set! v-errors '())
    (set! v-saved-lines '())
    (set! v-remset-trace '()))

  (define (clear-runs)
    (set! v-begin-system #f)
    (set! v-tags '())
    (set! v-settings '())
    (set! v-benchmark-name #f)
    (set! v-runs '()))

  (call-with-input-file filename
    (lambda (in)

      (define (read-tokens l)
        (let ((in (open-input-string l)))
          (let loop ((t '()))
            (let ((token (read in)))
              (if (eof-object? token)
                  (reverse t)
                  (loop (cons token t)))))))

      (define (read-one l)
        (read (open-input-string l)))

      (define (chop s prefix)
        (substring s (string-length prefix) (string-length s)))

      (define (butfirst l) (cdr (read-tokens l)))

      (define (match? l s) (eqv? (substring-match l s) 0))

      ; Expected order:
      ;  begin-system
      ;  tags
      ;  settings
      ;  <repeat>
      ;    benchmark-name
      ;    <repeat>
      ;      run-number
      ;      benchmark-command
      ;      <repeat>
      ;        stats-output
      ;      ps-output
      ;
      ; with warnings, errors, and saved-command-line anywhere.

      (define (get-line)
        (let ((l (read-line in)))
          (set! line-number (+ line-number 1))
          (cond ((or (eof-object? l)
                     (match? l "STATS-OUTPUT")
                     (match? l "RAW-MEMSTATS-BEFORE")
                     (match? l "RAW-MEMSTATS-AFTER")
                     (match? l "REMSET-PROFILE")
                     (match? l "REMSET-TRACE")
                     (match? l "WARNING")
                     (match? l "ERROR")
                     (match? l "Allocation failed")
                     (match? l "Larceny Panic")
                     (match? l "BENCHMARK-NAME")
                     (match? l "RUN-NUMBER")
                     (match? l "PS-OUTPUT")
                     (match? l "SETTING")
                     (match? l "TAG-")
                     (match? l "BENCHMARK-COMMAND")
                     (match? l "BEGIN-SYSTEM")
                     (match? l "SAVED-COMMAND-LINE"))
                 l)
                ((and (not (null? options))
                      (some? (lambda (x) 
                               (and (match? l (car x))
                                    (cdr x)))
                             options))
                 =>
                 (lambda (proc)
                   (let ((n (proc l)))
                     (if n
                         n
                         (get-line)))))
                (else
                 (get-line)))))

      (define (goto state . rest)
        (define (show l)
          (display "Line ")
          (display line-number)
          (display " >>> ")
          (display l) 
          (newline))
        (let ((l (get-line)))
          (cond ((eof-object? l)
                 (save-run)
                 (save-runs))
                ((match? l "WARNING") 
                 (show l)
                 (set! v-warnings (cons (chop l "WARNING ") v-warnings))
                 (apply goto state rest))
                ((match? l "ERROR") 
                 (show l)
                 (set! v-errors (cons (chop l "ERROR ") v-errors))
                 (apply goto state rest))
                ((or (match? l "Allocation failed")
                     (match? l "Larceny Panic"))
                 (show l)
                 (apply goto state rest))
                ((match? l "SAVED-COMMAND-LINE")
                 (set! v-saved-lines (cons (chop l "SAVED-COMMAND-LINE ") 
                                           v-saved-lines))
                 (apply goto state rest))
                ((match? l "REMSET-TRACE")
                 ; This only works when the remset trace data is on a
                 ; single line.  In some collectors it's not, and they
                 ; must be fixed!  See the code in the ROF collector
                 ; for an example of how to do it correctly.
                 (set! v-remset-trace (cons (read-one (chop l "REMSET-TRACE"))
                                            v-remset-trace))
                 (apply goto state rest))
                (else
                 (let loop ((rest rest))
                   (cond ((null? rest)
                          (set! errors (+ errors 1))
                          (display "ERROR: Failed to match \"")
                          (display l)
                          (display "\" in state ")
                          (display state)
                          (display " on line ")
                          (display line-number)
                          (newline)
                          (display "Attempting to sync...")
                          (newline)
                          (let loop ((l (get-line)))
                            (cond ((eof-object? l)
                                   (display "Failed to sync.")
                                   (newline))
                                  ((match? l "BEGIN-SYSTEM")
                                   (display "Synced on BEGIN-SYSTEM")
                                   (newline)
                                   (clear-run)
                                   (begin-system l))
                                  ((match? l "BENCHMARK-COMMAND")
                                   (display "Synced on BENCHMARK-COMMAND")
                                   (newline)
                                   (clear-run)
                                   (benchmark-command l))
                                  (else (loop (get-line))))))
                         ((match? l (car rest))
                          ((cadr rest) l))
                         (else
                          (loop (cddr rest)))))))))

      (define (save-run)
        (if v-run-number
            (set! v-runs (append v-runs 
                                 (list (make-run v-run-number 
                                                 v-benchmark-command 
                                                 v-stats-output 
                                                 v-ps-output
                                                 (reverse v-warnings)
                                                 (reverse v-errors)
                                                 (reverse v-saved-lines)
                                                 v-remset-profiles
                                                 (reverse v-remset-trace))))))
        (clear-run))

      (define (save-runs . rest)
        (if (not (null? v-runs))
            (set! v-systems (append v-systems 
                                    (list (make-system v-begin-system
                                                       v-tags
                                                       v-settings
                                                       v-benchmark-name
                                                       v-runs)))))
        
        (if (null? rest)
            (clear-runs)
            (set! v-runs '())))

      (define (begin-system l)

        (define (fix-system x)
          (let ((sz (caddr x)))         ; heap size or load factor
            (cond ((number? sz))        ; OK (unencoded, or load factor)
                  ((symbol? sz)
                   (let* ((s (symbol->string sz))
                          (c (char-upcase 
                              (string-ref s (- (string-length s) 1))))
                          (p (substring s 0 (- (string-length s) 1))))
                     (cond ((char=? c #\M)
                            (set-car! (cddr x) (* 1024 1024
                                                  (string->number p))))
                           ((char=? c #\K)
                            (set-car! (cddr x) (* 1024
                                                  (string->number p))))
                           (else
                            ???))))
                  (else
                   ???)))
          x)

        (save-run)
        (save-runs)
        (set! v-begin-system (fix-system (butfirst l)))
        (goto 'begin-system 
              "TAG-" tag))

      (define (tag l)
        (define (parse-tag)
          (cons (cond ((match? l "TAG-IDENTIFIER") 'identifier)
                      ((match? l "TAG-DATE") 'date)
                      ((match? l "TAG-HOSTNAME") 'hostname)
                      (else ???))
                (butfirst l)))
        (set! v-tags (cons (parse-tag) v-tags))
        (goto 'tag
              "SETTING" setting
              "TAG-" tag))

      (define (setting l)
        (set! v-settings (cons (butfirst l) v-settings))
        (goto 'setting
              "SETTING" setting
              "BENCHMARK-NAME" benchmark-name))
                                                     
      (define (benchmark-name l)
        (set! v-benchmark-name (butfirst l))
        (goto 'benchmark-name 
              "RUN-NUMBER" run-number))

      (define (run-number l)
        (save-run)
        (set! v-run-number (butfirst l))
        (goto 'run-number
              "BENCHMARK-COMMAND" benchmark-command))

      (define (benchmark-command l)
        (set! v-benchmark-command (chop l "BENCHMARK-COMMAND "))
        (goto 'benchmark-command
              "REMSET-PROFILE" remset-profile-before
              "STATS-OUTPUT" stats-output
              "RAW-MEMSTATS-BEFORE" raw-memstats-before
              "RUN-NUMBER" run-number         ; if failed but more runs
              "BENCHMARK-NAME" (lambda (l)    ; if failed but more benchmarks
                                 (save-run)
                                 (save-runs #f)
                                 (benchmark-name l))
              "BEGIN-SYSTEM" begin-system))   ; if failed and last run

      (define (remset-profile-before l)
        (set! v-remset-profile-tmp (read-one (chop l "REMSET-PROFILE ")))
        (goto 'remset-profile-before
              "STATS-OUTPUT" stats-output            ; success
              "RAW-MEMSTATS-BEFORE" raw-memstats-before
              "RUN-NUMBER" run-number                ; next run after failure
              "BENCHMARK-NAME" (lambda (l)           ; failed; more benchmarks
                                 (save-run)
                                 (save-runs #f)
                                 (benchmark-name l))
              "BEGIN-SYSTEM" begin-system))          ; next system after fail

      (define (remset-profile-after l)
        (let ((p (read-one (chop l "REMSET-PROFILE ")))
              (q v-remset-profile-tmp))
          (set! v-remset-profiles
                (cons (list (- (cadr p) (cadr q))
                            (- (cadddr p) (cadddr q)))
                      v-remset-profiles)))
        (goto 'remset-profile-after
              "REMSET-PROFILE" remset-profile-before
              "RUN-NUMBER" run-number                ; next run after failure
              "BENCHMARK-NAME" (lambda (l)           ; failed; more benchmarks
                                 (save-run)
                                 (save-runs #f)
                                 (benchmark-name l))
              "BEGIN-SYSTEM" begin-system            ; next system after fail
              "PS-OUTPUT" ps-output))                ; success

      (define (stats-output l)          ; Older-style output -- no longer used
        (set! v-stats-output 
              (append 
               v-stats-output
               (let ((data (read-one (chop l "STATS-OUTPUT "))))
                 (patch-stats data)
                 (let-values (((ssb add entries scanned has-np)
                               (crunch-remembered-sets data))
                              ((total array remembered imm xgen trans)
                               (crunch-simulated-barrier data))
                              ((cpu-gc cpu-promote cpu-collect cpu-full)
                               (crunch-gc-cpu-times data))
                              ((full-gc-time full-gcs)
                               (crunch-full-gc data)))
                   (list (append 
                          data
                          `((remset-ssb-transactions . ,ssb)
                            (remset-add-transactions . ,add)
                            (remset-entries-scanned . ,entries)
                            (remset-bytes-scanned . ,scanned)
                            (elapsed-time-full-gc . ,full-gc-time)
                            (count-full-gc . ,full-gcs)
                            (cpu-time-gc . ,cpu-gc)
                            (cpu-time-gc-promotion . ,cpu-promote)
                            (cpu-time-gc-collection . ,cpu-collect)
                            (cpu-time-gc-full . ,cpu-full)
                            (has-np-remset . ,has-np)
                            (swb-total-assignments . ,total)
                            (swb-array-assignments . ,array)
                            (swb-filtered-remembered . ,remembered)
                            (swb-filtered-constant . ,imm)
                            (swb-filtered-not-old-to-young . ,xgen)
                            (swb-transactions . ,trans))))))))
        (goto 'stats-output
              "STATS-OUTPUT" stats-output            ; next iteration
              "RUN-NUMBER" run-number                ; next run after failure
              "BENCHMARK-NAME" (lambda (l)           ; failed; more benchmarks
                                 (save-run)
                                 (save-runs #f)
                                 (benchmark-name l))
              "BEGIN-SYSTEM" begin-system            ; next system after fail
              "PS-OUTPUT" ps-output                  ; success
              "REMSET-PROFILE" remset-profile-after  ; success
              ))

      (define (raw-memstats-before l)
        (set! v-raw-before-tmp (read-one (chop l "RAW-MEMSTATS-BEFORE ")))
        (goto 'raw-memstats-before
              "RAW-MEMSTATS-AFTER" raw-memstats-after  ; success
              "RUN-NUMBER" run-number                ; next run after failure
              "BENCHMARK-NAME" (lambda (l)           ; failed; more benchmarks
                                 (save-run)
                                 (save-runs #f)
                                 (benchmark-name l))
              "BEGIN-SYSTEM" begin-system))          ; next system after fail

      (define (raw-memstats-after l)
        (set! v-stats-output
              (append 
               v-stats-output
               (list (cons v-raw-before-tmp
                           (read-one (chop l "RAW-MEMSTATS-AFTER "))))))
        (goto 'raw-memstats-after
              "RAW-MEMSTATS-BEFORE" raw-memstats-before  ; Next iteration
              "RUN-NUMBER" run-number                 ; next run after fail
              "BENCHMARK-NAME" (lambda (l)            ; failed, more benchmarks
                                 (save-run)
                                 (save-runs #f)
                                 (benchmark-name l))
              "BEGIN-SYSTEM" begin-system             ; next system after fail
              "PS-OUTPUT" ps-output                   ; success
              "RESMET-PROFILE" remset-profile-after   ; success
              ))
              
      (define (ps-output l)
        (set! v-ps-output (read-one (chop l "PS-OUTPUT")))
        (goto 'ps-output
              "RUN-NUMBER" run-number                 ; more runs
              "BENCHMARK-NAME" (lambda (l)            ; more benchmarks
                                 (save-run)
                                 (save-runs #f)
                                 (benchmark-name l))
              "BEGIN-SYSTEM" begin-system))           ; next major system
        
      (goto 'start "BEGIN-SYSTEM" begin-system)
      (cond ((zero? errors))
            ((= errors 1) (display "One error.") (newline))
            (else (display errors) (display " errors.") (newline)))
      v-systems)))


(define (load-crunched-data crunched-file)
  (format #t "Loading ~a~%" crunched-file)
  (let ((c (call-with-input-file crunched-file read)))

    ; Traverse the structure and replace every pair of raw stats with 
    ; computed data.  Structure is modified in place.

    (for-each 
     (lambda (system)
       (let ((gctype (car (system.info system))))
         (for-each 
          (lambda (run)
            (for-each
             (lambda (iteration)
               (let ((data (produce-data-from-raw-stats
                            (car iteration)
                            (cdr iteration)
                            gctype)))
                 (set-car! iteration (car data))
                 (set-cdr! iteration (cdr data))))
             (run.stats run)))
          (system.runs system))))
     c)

    ; Return the structure

    c))


; This operates on the crunched representation.  It destructively adds the
; options given to the end of the system.info field in the representation.  
; It returns the data.

(define (crunched-add-option! data . options)
  (for-each (lambda (system)
              (let ((i (system.info system)))
                (append! i (list-copy options))))
            data)
  data)



; For newer-style truly raw stats data

(define (vref n)
  (lambda (v)
    (vector-ref v n)))

(define (cvref n)
  (lambda (v)
    (if (< n (vector-length v))
        (vector-ref v n)
        0)))

(define (vref-check n tag)
  (lambda (v)
    (let ((x (vector-ref v n)))
      (if (and (> (vector-length x) 0)
               (not (eq? (vector-ref x 0) tag)))
          (error "Failed to match tag " tag " with " x))
      x)))

(define (vref-check-all n tag)
  (lambda (v)
    (let ((x (vector-ref v n)))
      (do ((i 0 (+ i 1)))
          ((= i (vector-length x)))
        (let ((y (vector-ref x i)))
          (if (and (> (vector-length y) 0)
                   (not (eq? (vector-ref y 0) tag)))
              (error "Failed to match tag " tag " with " y))))
      x)))

(define bigref 
  (let ((two^29 (expt 2 29)))
    (lambda (n) 
      (lambda (v)
        (+ (* (vector-ref v n) two^29)
           (vector-ref v (+ n 1)))))))

; Main accessors
(define raw-version (vref 0))
(define raw-system-v (vref-check 1 'system_overall))
(define raw-gc-v (vref-check 2 'gc_memstat_t))
(define raw-gclib-v (vref-check 3 'gclib_memstat_t))
(define raw-generations-v (vref-check-all 4 'gen_memstat_t))
(define raw-remsets-v (vref-check-all 5 'remset_memstat_t))
(define raw-stack-v (vref-check 6 'stack_memstat_t))
(define raw-swb-v (vref-check 7 'swb_memstat_t))
(define raw-gc-event-v (vref-check 8 'gc_event_memstat_t))

; system overall
(define raw-system-ms-real (vref 1))
(define raw-system-ms-system (vref 2))
(define raw-system-ms-user (vref 3))
(define raw-system-minor-faults (vref 4))
(define raw-system-major-faults (vref 5))

; gc_memstat_t
(define raw-gc-allocated (bigref 1))
(define raw-gc-reclaimed (bigref 3))
(define raw-gc-objects-copied (bigref 5))
(define raw-gc-words-copied (bigref 7))
(define raw-gc-objects-moved (bigref 9))
(define raw-gc-words-moved (bigref 11))
(define raw-gc-np-k (vref 13))
(define raw-gc-np-j (vref 14))
(define raw-gc-full-collections (vref 15))
(define raw-gc-full-ms-collection (vref 16))
(define raw-gc-full-ms-collection-cpu (vref 17))
(define raw-gc-full-objects-marked (bigref 18))
(define raw-gc-full-words-marked (bigref 20))
(define raw-gc-full-pointers-traced (bigref 22))

; gclib_memstat_t
(define raw-gclib-heap-allocated (vref 1))
(define raw-gclib-heap-allocated-max (vref 2))
(define raw-gclib-remset-allocated (vref 3))
(define raw-gclib-remset-allocated-max (vref 4))
(define raw-gclib-rts-allocated (vref 5))
(define raw-gclib-rts-allocated-max (vref 6))
(define raw-gclib-heap-fragmentation (vref 7))
(define raw-gclib-heap-fragmentation-max (vref 8))
(define raw-gclib-mem-allocated (vref 9))
(define raw-gclib-mem-allocated-max (vref 10))

; gen_memstat_t
(define raw-gen-major-id (vref 1))
(define raw-gen-minor-id (vref 2))
(define raw-gen-target (vref 3))
(define raw-gen-allocated (vref 4))
(define raw-gen-used (vref 5))
(define raw-gen-promotions (vref 6))
(define raw-gen-collections (vref 7))
(define raw-gen-ms-promotion (vref 8))
(define raw-gen-ms-promotion-cpu (vref 9))
(define raw-gen-ms-collection (vref 10))
(define raw-gen-ms-collection-cpu (vref 11))

; remset_memstat_t
(define raw-remset-major-id (vref 1))
(define raw-remset-minor-id (vref 2))
(define raw-remset-allocated (vref 3))
(define raw-remset-max-allocated (vref 4))
(define raw-remset-allocated-max raw-remset-max-allocated) ; Alias
(define raw-remset-used (vref 5))
(define raw-remset-live (vref 6))
(define raw-remset-ssb-recorded (bigref 7))
(define raw-remset-recorded (bigref 9))
(define raw-remset-objs-scanned (bigref 11))
(define raw-remset-words-scanned (bigref 13))
(define raw-remset-removed (bigref 15))
(define raw-remset-cleared (vref 17))
(define raw-remset-scanned (vref 18))
(define raw-remset-compacted (vref 19))

; stack_memstat_t
(define raw-stack-stacks-created (vref 1))
(define raw-stack-words-flushed (bigref 2))
(define raw-stack-frames-flushed (bigref 4))
(define raw-stack-frames-restored (bigref 6))

; gc_event_memstat_t
(define raw-gc-event-gctime (bigref 1))
(define raw-gc-event-promtime (bigref 3))
(define raw-gc-event-free-unused (bigref 5))
(define raw-gc-event-root-scan-gc (bigref 7))
(define raw-gc-event-root-scan-prom (bigref 9))
(define raw-gc-event-los-sweep-gc (bigref 11))
(define raw-gc-event-los-sweep-prom (bigref 13))
(define raw-gc-event-remset-scan-gc (bigref 15))
(define raw-gc-event-remset-scan-prom (bigref 17))
(define raw-gc-event-tospace-scan-gc (bigref 19))
(define raw-gc-event-tospace-scan-prom (bigref 21))
(define raw-gc-event-reset-after-gc (bigref 23))
(define raw-gc-event-decrement-after-gc (bigref 25))
(define raw-gc-event-dof-remset-scan (bigref 27))
(define raw-gc-event-sweep-shadow (bigref 29))
(define raw-gc-event-msgc-mark (bigref 31))
(define raw-gc-event-sweep-dof-sets (bigref 33))
(define raw-gc-event-sweep-remset (bigref 35))
(define raw-gc-event-sweep-los (bigref 37))
(define raw-gc-event-assimilate-prom (bigref 39))
(define raw-gc-event-assimilate-gc (bigref 41))
(define raw-gc-event-copied-by-gc (vref 43))
(define raw-gc-event-copied-by-prom (vref 44))
(define raw-gc-event-words-forwarded (vref 45))
(define raw-gc-event-ptrs-forwarded (vref 46))
(define raw-gc-event-gc-barrier-hit (vref 47))
(define raw-gc-event-remset-large-objs-scanned (vref 48))
(define raw-gc-event-remset-large-obj-words-scanned (vref 49))
(define raw-gc-event-moved-by-gc (cvref 50))
(define raw-gc-event-moved-by-prom (cvref 51))

; Does not include raw data vectors or simulated barrier stats.
;
; Def: a _major_ collection is any collection involving the dynamic area
;      that reclaims storage in the dynamic area.

; The information about the heap structure is not directly available in
; the raw stats, but it's easy to infer it if we're given information
; about the type of collector (sc,gen,rof,dof).  For the SC collector,
; every collection is a major collection.  For the GEN collector, the
; next-to-oldest generation is the sole dynamic generation.  For the 
; ROF and DOF collectors, all generations except the youngest and oldest 
; are in the dynamic area -- in my current benchmark setup.  Good enough.

(define (produce-data-from-raw-stats before after gctype)

  (define (gc-count g)
    (+ (raw-gen-collections g) (raw-gen-promotions g)))

  (define (sum fn l)
    (apply + (map fn l)))

  (define (diff-list fn after before)
    (map (lambda (a b) (- (fn a) (fn b)))
         after
         before))

  (define (minor-generations n)         ; Returns list of minor generations
    (case gctype
      ((sc) '())
      ((gen) (case n
               ((3) '(0))               ; 2 generations + static
               ((4) '(0 1))             ; 3 generations + static
               (else ???)))
      ((rof np) '(0))                   ; ROF and NP are the same
      (else 
       (if (and (pair? gctype)          ; DOF type is (DOF k)
                (eq? (car gctype) 'dof))
           '(0)
           (error "Bogus GC type " gctype)))))

  (define (major-generations n)         ; Returns list of major generations
    (let ((min (minor-generations n)))
      (filter (lambda (maj)
                (not (memv maj min)))
              (iota (- n 1)))))         ; ignore static area

  (if (not (and (= 0.49 (raw-version before))
                (= 0.49 (raw-version after))))
      (begin
        (display "WARNING: Expected raw stats version 0.49, got ")
        (display (raw-version before))
        (display " and ")
        (display (raw-version after))
        (newline)
        (newline)))

  (let* ((gc-v-before          (raw-gc-v before))
         (gc-v-after           (raw-gc-v after))
         (system-v-before      (raw-system-v before))
         (system-v-after       (raw-system-v after))
         (gclib-v-after        (raw-gclib-v after))
         (generations-l-before (vector->list (raw-generations-v before)))
         (generations-l-after  (vector->list (raw-generations-v after)))
         (remsets-l-before     (vector->list (raw-remsets-v before)))
         (remsets-l-after      (vector->list (raw-remsets-v after)))
         (stack-after          (raw-stack-v after))
         (stack-before         (raw-stack-v before))
         (major                (major-generations 
                                (length generations-l-after)))
         (minor                (minor-generations 
                                (length generations-l-after)))
         (event-v-before       (raw-gc-event-v before))
         (event-v-after        (raw-gc-event-v after))

         (frames-restored
          (- (raw-stack-frames-restored stack-after)
             (raw-stack-frames-restored stack-before)))
         (frames-flushed
          (- (raw-stack-frames-flushed stack-after)
             (raw-stack-frames-flushed stack-before)))
         (bytes-allocated
          (* 4 (- (raw-gc-allocated gc-v-after)
                  (raw-gc-allocated gc-v-before))))
         (bytes-reclaimed
          (* 4 (- (raw-gc-reclaimed gc-v-after)
                  (raw-gc-reclaimed gc-v-before))))
         (bytes-marked                  ; Total
          (+ (* 4 (- (raw-gc-words-copied gc-v-after)
                     (raw-gc-words-copied gc-v-before)))
             (* 4 (- (raw-gc-words-moved gc-v-after)
                     (raw-gc-words-moved gc-v-before)))
             (* 4 (- (raw-gc-full-words-marked gc-v-after)
                     (raw-gc-full-words-marked gc-v-before)))))
         (bytes-copied
          (* 4 (- (raw-gc-words-copied gc-v-after)
                  (raw-gc-words-copied gc-v-before))))
         (bytes-copied-promotion 
          (* 4 (- (raw-gc-event-copied-by-prom event-v-after)
                  (raw-gc-event-copied-by-prom event-v-before))))
         (bytes-copied-collection 
          (* 4 (- (raw-gc-event-copied-by-gc event-v-after)
                  (raw-gc-event-copied-by-gc event-v-before))))
         (bytes-moved-promotion 
          (* 4 (- (raw-gc-event-moved-by-prom event-v-after)
                  (raw-gc-event-moved-by-prom event-v-before))))
         (bytes-moved-collection 
          (* 4 (- (raw-gc-event-moved-by-gc event-v-after)
                  (raw-gc-event-moved-by-gc event-v-before))))
         (bytes-moved 
          (* 4 (- (raw-gc-words-moved gc-v-after)
                  (raw-gc-words-moved gc-v-before))))
         (bytes-marked-full-gc 
          (* 4 (- (raw-gc-full-words-marked gc-v-after)
                  (raw-gc-full-words-marked gc-v-before))))
         (elapsed-time
          (- (raw-system-ms-real system-v-after)
             (raw-system-ms-real system-v-before)))
         (cpu-time-user
          (- (raw-system-ms-user system-v-after)
             (raw-system-ms-user system-v-before)))
         (cpu-time-system
          (- (raw-system-ms-system system-v-after)
             (raw-system-ms-system system-v-before)))
         (elapsed-time-major-gc 0)      ; for now
         (elapsed-time-minor-gc 0)      ; for now
         (elapsed-time-full-gc 
          (- (raw-gc-full-ms-collection gc-v-after) 
             (raw-gc-full-ms-collection gc-v-before)))
         (elapsed-time-gc 0)            ; for now
         (cpu-time-gc-promotion 
          (- (sum raw-gen-ms-promotion-cpu generations-l-after)
             (sum raw-gen-ms-promotion-cpu generations-l-before)))
         (cpu-time-gc-collection 
          (- (sum raw-gen-ms-collection-cpu generations-l-after)
             (sum raw-gen-ms-collection-cpu generations-l-before)))
         (cpu-time-gc-full 
          (- (raw-gc-full-ms-collection-cpu gc-v-after)
             (raw-gc-full-ms-collection-cpu gc-v-before)))
         (cpu-time-gc
          (+ cpu-time-gc-promotion cpu-time-gc-collection cpu-time-gc-full))
         (cpu-time-gc-minor             ; Note, does not include promotions!
          (- (sum raw-gen-ms-collection-cpu 
                  (map (lambda (n) (list-ref generations-l-after n)) 
                       minor))
             (sum raw-gen-ms-collection-cpu 
                  (map (lambda (n) (list-ref generations-l-before n)) 
                       minor))))
         (cpu-time-gc-major
          (- (sum raw-gen-ms-collection-cpu 
                  (map (lambda (n) (list-ref generations-l-after n))
                       major))
             (sum raw-gen-ms-collection-cpu 
                  (map (lambda (n) (list-ref generations-l-before n))
                       major))))
         (count-gc                      ; Not including full collections!
          (- (sum gc-count generations-l-after)
             (sum gc-count generations-l-before)))
         (count-full-gc
          (- (raw-gc-full-collections gc-v-after)
             (raw-gc-full-collections gc-v-before)))
         (count-minor-gc                ; Note, does not include promotions!
          (- (sum raw-gen-collections 
                  (map (lambda (n) (list-ref generations-l-after n)) 
                       minor))
             (sum raw-gen-collections 
                  (map (lambda (n) (list-ref generations-l-before n)) 
                       minor))))
         (count-major-gc
          (- (sum raw-gen-collections 
                  (map (lambda (n) (list-ref generations-l-after n))
                       major))
             (sum raw-gen-collections 
                  (map (lambda (n) (list-ref generations-l-before n))
                       major))))
         (bytes-max-heap-used 
          (* 4 (raw-gclib-heap-allocated-max gclib-v-after)))
         (bytes-max-remset-used
          ; The field gclib-remset-allocated-max is valid, if not zero.
          ; If zero, then data for remembered-set allocation have been
          ; folded into data for RTS allocation. (This happens when the
          ; page table is preallocated for the entire 4GB range because
          ; each page entry is only 1 byte and does not contain
          ; information about the RTS allocation subtype.)  In that case,
          ; remset allocation is _roughly_ the max RTS allocation minus 
          ; one megabyte, the size of the page table.  
          ;
          ; HOWEVER, if the DOF collector is run and uses the full collector
          ; then all bets are off -- the field contains also allocation for
          ; the mark bitmap and the mark stack.  So it's best to set the
          ; system up to measure the remembered-set allocation accurately.
          (let ((x (raw-gclib-remset-allocated-max gclib-v-after)))
            (if (not (zero? x))
                (* 4 x)
                (- (* 4 (raw-gclib-rts-allocated-max gclib-v-after))
                   (* 1024 1024)))))
         (remset-entries-scanned
          (diff-list raw-remset-objs-scanned remsets-l-after remsets-l-before))
         (remset-entries-removed
          (diff-list raw-remset-removed remsets-l-after remsets-l-before))
         (remset-bytes-scanned
          (map (lambda (x) (* 4 x))
               (diff-list raw-remset-words-scanned
                          remsets-l-after remsets-l-before)))
         (remset-ssb-transactions
          (diff-list raw-remset-ssb-recorded remsets-l-after remsets-l-before))
         (remset-add-transactions
          (diff-list raw-remset-recorded remsets-l-after remsets-l-before))
         (has-np-remset
          (> (raw-gc-np-k gc-v-after) 0)))

    `((bytes-allocated . ,bytes-allocated)
      (bytes-reclaimed . ,bytes-reclaimed)
      (bytes-marked . ,bytes-marked)
      (bytes-copied . ,bytes-copied)
      (bytes-copied-promotion . ,bytes-copied-promotion)
      (bytes-copied-collection . ,bytes-copied-collection)
      (bytes-moved-promotion . ,bytes-moved-promotion)
      (bytes-moved-collection . ,bytes-moved-collection)
      (bytes-moved . ,bytes-moved)
      (bytes-marked-full-gc . ,bytes-marked-full-gc)
      (elapsed-time . ,elapsed-time)
      (cpu-time-user . ,cpu-time-user)
      (cpu-time-system . ,cpu-time-system)
      (elapsed-time-gc . ,elapsed-time-gc)
      (elapsed-time-major-gc . ,elapsed-time-major-gc)
      (elapsed-time-minor-gc . ,elapsed-time-minor-gc)
      (elapsed-time-full-gc . ,elapsed-time-full-gc)
      (cpu-time-gc . ,cpu-time-gc)
      (cpu-time-gc-promotion . ,cpu-time-gc-promotion)
      (cpu-time-gc-collection . ,cpu-time-gc-collection)
      (cpu-time-gc-minor . ,cpu-time-gc-minor)
      (cpu-time-gc-major . ,cpu-time-gc-major)
      (cpu-time-gc-full . ,cpu-time-gc-full)
      (count-gc . ,count-gc)
      (count-full-gc . ,count-full-gc)
      (bytes-max-heap-used . ,bytes-max-heap-used)
      (bytes-max-remset-used . ,bytes-max-remset-used)
      (remset-entries-scanned . ,remset-entries-scanned)
      (remset-entries-removed . ,remset-entries-removed)
      (remset-bytes-scanned . ,remset-bytes-scanned)
      (remset-ssb-transactions . ,remset-ssb-transactions)
      (remset-add-transactions . ,remset-add-transactions)
      (count-minor-gc . ,count-minor-gc)
      (count-major-gc . ,count-major-gc)
      (has-np-remset . ,has-np-remset)
      (frames-flushed . ,frames-flushed)
      (frames-restored . ,frames-restored))))


; The following code supports old-style STATS-OUTPUT data.

; NOTE: note the following problem.  According to the Manual the NP
; remembered set entry (entry 26) is supposed to be #f if the collector
; is not the NP collector, but it is not.  Instead it is always a vector
; containing a single subvector (this is also an error -- it should be a
; flat vector).
;
; To tell whether the NP collector is being used or not it is therefore
; necessary to examine the flag fields (4 and 5) of the _generation_
; statistics.  If those flags are not set for any generation, we are not
; using the NP collector; otherwise we are, and the data in entry 26 are
; valid.  
;
; Code that uses entry 26 should be able to: 
;
;  - deal with it if it is #f (because I'll fix the bug eventually) 
;  - deal with it if it is a nested vector 
;  - deal with it if it is a flat vector (because I'll fix the bug eventually)
;
; Larceny v0.44 has the same bug.

(define (crunch-remembered-sets the-data)

  (define (process-field then now field mpy)
    (map (lambda (before after)
           (* mpy (- (vector-ref after field) (vector-ref before field))))
         then
         now))

  (if (not (assq 'raw-data-before the-data))
      (values '() '() '() '() #f)
      (let* ((raw-before (cdr (assq 'raw-data-before the-data)))
             (raw-after  (cdr (assq 'raw-data-after the-data)))
             (then       (vector->list (vector-ref raw-before 8)))
             (now        (vector->list (vector-ref raw-after 8)))
             (l          (length then)))
        (let ((ssb-stats     (process-field then now 8 1))
              (add-stats     (process-field then now 4 1))
              (entries-stats (process-field then now 6 1))
              (scanned-stats (process-field then now 7 4)))
          (values ssb-stats
                  add-stats
                  entries-stats
                  scanned-stats
                  #f)))))

; Returns cpu-tot cpu-prom cpu-collect cpu-full

; These accessors are present in v0.48 and later.

(define (memstats-gc-total-cpu-time v) (vector-ref v 44))
(define (memstats-gc-promotion-cpu-time v) (vector-ref v 45))
(define (memstats-fullgc-cpu-time v) (vector-ref v 43))

(define (crunch-gc-cpu-times the-data)
  (let ((raw-before (cdr (assq 'raw-data-before the-data)))
        (raw-after  (cdr (assq 'raw-data-after the-data))))
    (values (- (memstats-gc-total-cpu-time raw-after)
               (memstats-gc-total-cpu-time raw-before))
            (- (memstats-gc-promotion-cpu-time raw-after)
               (memstats-gc-promotion-cpu-time raw-before))
            (- (- (memstats-gc-total-cpu-time raw-after)
                  (memstats-gc-promotion-cpu-time raw-after)
                  (memstats-fullgc-cpu-time raw-after))
               (- (memstats-gc-total-cpu-time raw-before)
                  (memstats-gc-promotion-cpu-time raw-before)
                  (memstats-fullgc-cpu-time raw-before)))
            (- (memstats-fullgc-cpu-time raw-after) 
               (memstats-fullgc-cpu-time raw-before)))))

; returns time and number of collections

(define (crunch-full-gc the-data)
  (let* ((raw-before (cdr (assq 'raw-data-before the-data)))
         (raw-after  (cdr (assq 'raw-data-after the-data))))
    (values (- (vector-ref raw-after 35)
               (vector-ref raw-before 35))
            (- (vector-ref raw-after 34)
               (vector-ref raw-before 34)))))

(define (crunch-simulated-barrier the-data)

  (define (swb-data raw)
    (list (if (>= (vector-length raw) 30)
              (vector-ref raw 29)       ; total assignments
              0)                        ; old memstats format
          (vector-ref raw 16)           ; array/procedure assignments
          (vector-ref raw 17)           ; filtered: already remembered
          (vector-ref raw 18)           ; filtered: not a pointer
          (vector-ref raw 19)           ; filtered: not old-to-young
          (vector-ref raw 20)))         ; transactions

  (if (not (assq 'raw-data-before the-data))
      (values 0 0 0 0 0 0)
      (let ((then (swb-data (cdr (assq 'raw-data-before the-data))))
            (now  (swb-data (cdr (assq 'raw-data-after the-data)))))
        (apply values 
               (map (lambda (now then)
                      (if (< now then)
                          (begin
                            (display "WARNING: adjusting for overflow in SWB data.")
                            (newline)
                            (+ (expt 2 29) (- now then)))
                          (- now then)))
                    now
                    then)))))

; Some postprocessing to cover up for bugs.
; Override this when you want to patch.
; Works by side effect only.

(define (patch-stats stats)
  #f)

; eof
