; Copyright 1998 Lars T Hansen.
;
; $Id: makefile.sch,v 1.26 1999/11/18 16:32:11 lth Exp $
;
; Larceny development system -- makefile for compiling Scheme files.
;
; Procedures to call:
;  make-sparc-heap
;  make-petit-heap
;  make-extended-petit-heap
;  make-auxlib
;  make-petit-auxlib
;  make-compat
;  make-petit-compat
;  make-compiler
;  make-petit-compiler
;  make-sparcasm
;  make-petitasm
;  make-petit-petitasm
;  make-gc-testsuite
;  make-regression-test
;  make-development-environment
;     makes auxlib, compat, compiler, sparcasm, petitasm
;  make-petit-development-environment
;  make-petit-heap-with-compiler

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Action procedures for the rules.

(define (make-compile target deps) 
  (display "Compiling ") (display target) (newline)
  (compile313 (car deps) target))

(define (make-assemble target deps)
  (display "Assembling ") (display target) (newline)
  (assemble313 (car deps) target))

(define (make-compile-file target deps)
  (display "Making ") (display target) (newline)
  (compile-file (car deps)))

(define (make-compile-and-assemble target deps)
  (display "Making ") (display target) (newline)
  (compile-and-assemble313 (car deps) target))

(define (make-assemble-file target deps)
  (display "Making ") (display target) (newline)
  (assemble-file (car deps)))

(define (make-dumpheap target files)
  (display "Dumping ") (display target) (newline)
  (delete-file target)
  (build-heap-image target files))

(define *extended-files* '())

(define (make-copy target src)
  (display "Copying ") (display target) (newline)
  (call-with-input-file (car src)
    (lambda (inp)
      (delete-file target)
      (call-with-output-file target
        (lambda (outp)
          (let loop ((item (read-char inp)))
            (if (eof-object? item)
                #t
                (begin (write-char item outp)
                       (loop (read-char inp))))))))))

(define (objects path ext files . rest)
  (let ((substitutions (if (null? rest) '() (car rest))))
    (define (loop files)
      (cond ((null? files) 
             '())
            ((string? (car files))
             (cons (string-append path (car files) ext)
                   (loop (cdr files))))
            ((symbol? (car files))
             (let ((probe (assq (car files) substitutions)))
               (cond ((not probe)
                      (error "objects: No substitutions found for " 
                             (car files)))
                     ((not (cdr probe))
                      (loop (cdr files)))
                     (else
                      (cons (cdr probe) (loop (cdr files)))))))
            (else
             ???)))
    (loop files)))

(define (replace-extension ext files)
  (map (lambda (file)
         (rewrite-file-type file '(".sch" ".h") ext))
       files))

(define (machine-relative x)
  (string-append (nbuild-parameter 'machine-source) x))

(define (common-relative x)
  (string-append (nbuild-parameter 'common-source) x))

(define (auxlib-relative x)
  (string-append (nbuild-parameter 'auxiliary) x))

(define (common-endian x)
  (string-append (nbuild-parameter 'common-source) 
                 x
                 (if (eq? (nbuild-parameter 'endianness) 'little) 
                     "-el" 
                     "-be")
                 ".lop"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Projects for building the basic heap images.

; These are the base names of all the files that make up the basic heap.

(define common-heap-files
  '(
    ; Fundamental

    "malcode"           ; really basic things
    "typetags"          ; type tags
    "syscall-id"        ; syscall ID numbers
    osdep               ; OS dependent system interface (to be substituted!)
    "system-interface"  ; OS independent system interface
    "error0"            ; Boot-time 'error' procedure.
    primops             ; primop procedures (to be substituted!)

    ; General library

    "sysparam"          ; system parameters
    "struct"            ; structures
    "command-line"      ; command line arguments
    "list"              ; list procedures
    "vector"            ; vector procedures
    "string"            ; string and bytevector procs
    "control"           ; control procedures
    "hash"              ; hash functions
    "hashtable"         ; hashtables
    "preds"             ; some predicates
    "oblist"            ; symbol table
    "mcode"             ; millicode support
    "memstats"          ; runtime stats
    "ecodes"            ; exception codes
    "ehandler"          ; exception handler
    "error"             ; error/reset system
    "timer"             ; timer interrupts
    "exit"              ; exit procedure; exit/init hooks
    "dump"              ; dump-heap procedure
    "secret"            ; some "hidden" top-level names

    ; New I/O subsystem

    "iosys"             ; basic system
    "fileio"            ; file ports
    "conio"             ; console ports, i.e., terminal
    "stringio"          ; string ports
;    "transio"           ; transcript ports
    "stdio"             ; user-level procedures
    "print"             ; write/display
    "ioboot"            ; one-time initialization

    "format"            ; `format' procedure.
    "sort"              ; `sort' and `sort!' (used by new macro expander)
    "number"            ; arithmetic
    "globals"           ; `global' offsets (for memstats)

    ; It's important for bellerophon to be loaded as late as possible
    ; because it depends on much of the rest of the system.

    "profile"           ; Profiling code
    bignum-endian       ; Endian-dependent bignum support (to be substituted!)
    "bignums"           ; Bignum support
    "ratnums"           ; Ratnum support
    "rectnums"          ; Rectnum support
    flonum-endian       ; Endian-dependent flonum support (to be substituted!)
    "flonums"           ; Flonum support
    "contag"            ; Contagion
    "num2str"           ; Number printer
    "belle"             ; Algorithm bellerophon
    "str2num"           ; Number parser
    "reader"            ; Reader
    "env"               ; R5RS environments
    "procinfo"          ; Heuristic procedure information
    "load"              ; Loader
    "eval"              ; Eval procedure
    "syshooks"          ; System functions
    "gcctl"             ; Garbage collector policy control
    toplevel            ; top-level environment (to be substituted!)
    extra               ; any private libraries (to be substituted!)
    "go"                ; Driver

    ))

; Files in the bootstrap evaluator, repl, and toplevel env.

(define eval-files
  (append
   (nbuild-files 'repl-source '("main" "reploop"))
   (nbuild-files 'interp-source '("interp" "interp-prim" "switches"))
   (nbuild-files 'compiler 
                 '("pass1" "pass1.aux" "pass2.aux" "prefs" 
                   "syntaxenv" "syntaxrules" "lowlevel" "expand" "usual"))
   (nbuild-files 'interp-source '("macro-expand"))
   ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building the sparc-larceny heap image.

(define sparc-heap-project
  (let ((sparc-heap-files 
         (objects "Lib/Common/"
                  ".lop"
                  common-heap-files
                  '((primops . "Lib/Sparc/primops.lop")
                    (toplevel . "Lib/Sparc/toplevel.lop")
                    (flonum-endian . "Lib/Common/flonums-be.lop")
                    (bignum-endian . "Lib/Common/bignums-be.lop")
                    (osdep . "Lib/Common/sys-unix.lop")
                    (extra . #f))))
        (sparc-eval-files
         (objects "" ".lop" eval-files)))
    (make:project "sparc.heap"
      `(rules
        (".lop" ".mal" ,make-assemble)
        (".lop" ".lap" ,make-assemble)
        (".lap" ".sch" ,make-compile)
        (".sch" ".sh"  ,make-copy))
      `(targets 
        ("sparc.heap" ,make-dumpheap))
      `(dependencies                    ; Order matters.  [Why??!]
        ("sparc.heap" ,sparc-heap-files)
        ("sparc.heap" ,sparc-eval-files)
        ("Lib/Common/ecodes.sch" ,(nbuild-files 'build '("except.sh")))
        ("Lib/Common/globals.sch" ,(nbuild-files 'build '("globals.sh")))))))
     
(define (make-sparc-heap . rest)
  (make:pretend (not (null? rest)))
  (let ((iup (integrate-usual-procedures)))
    (dynamic-wind 
     (lambda () (integrate-usual-procedures #t))
     (lambda () (make:make sparc-heap-project "sparc.heap"))
     (lambda () (integrate-usual-procedures iup)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building the petit-larceny heap image.

(define petit-heap-files '())
(define petit-eval-files '())

(define (petit-select-target target)
  (define (select target)
    (set! petit-heap-files
	  (objects (nbuild-parameter 'common-source) 
		   ".lop" 
		   common-heap-files
		   `((primops  . ,(machine-relative "primops.lop"))
		     (toplevel . ,(machine-relative "toplevel.lop"))
		     (flonum-endian . ,(common-endian "flonums"))
		     (bignum-endian . ,(common-endian "bignums"))
		     (osdep    . ,(common-relative target))
		     (extra    . ,(machine-relative "loadable.lop")))))
    (set! petit-eval-files
	  (objects "" ".lop" eval-files))
    (unspecified))

  (case target
    ((unix) (select "sys-unix.lop"))
    ((macos) (select "sys-macos.lop"))
    ((help) 
     (display "Targets are unix, macos.")
     (newline))
    (else 
     (error "Unsupported target "
	    target 
	    "; try one of unix, macos."))))
 
(define (make-petit-heap-project heap-dumper)
  (make:project "petit.heap"
    `(rules 
      (".lop" ".mal" ,make-assemble)
      (".lop" ".lap" ,make-assemble)
      (".lap" ".sch" ,make-compile)
      (".sch" ".sh"  ,make-copy))
    `(targets
      ("petit.heap" ,heap-dumper))
    `(dependencies                    ; Order matters [why??!]
      ("petit.heap" ,petit-heap-files)
      ("petit.heap" ,petit-eval-files)
      (,(common-relative "ecodes.sch") ,(nbuild-files 'build '("except.sh")))
      (,(common-relative "globals.sch")
       ,(nbuild-files 'build '("globals.sh"))))))

(define (make-petit-heap . rest)
  (if (null? petit-heap-files)
      (petit-select-target (nbuild-parameter 'target-os)))
  (let ((petit-heap-project
	 (make-petit-heap-project make-dumpheap)))
    (make:pretend (not (null? rest)))
       (let ((iup (integrate-usual-procedures)))
	 (dynamic-wind 
	  (lambda () (integrate-usual-procedures #t))
	  (lambda () (make:make petit-heap-project "petit.heap"))
	  (lambda () (integrate-usual-procedures iup))))))

(define (make-petit-libclean)
  (let ((files (append petit-heap-files petit-eval-files)))
    (for-each (lambda (x)
                (if (file-exists? x)
                    (delete-file x)))
              (append files
                      (map (lambda (x) (rewrite-file-type x '(".lop") ".c"))
                           files)
                      (map (lambda (x) (rewrite-file-type x '(".lop") ".o"))
                           files)
                      (map (lambda (x) (rewrite-file-type x '(".lop") ".lap"))
                           files)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building all the files in the Compiler subdirectory.

(define (make-compiler-project file-type)
  (let ((compiler-files 
         (replace-extension file-type (nbuild:twobit-files)))
        (comp-util-files 
         (replace-extension file-type (nbuild:utility-files)))
        (other-util-files
         (objects (nbuild-parameter 'util)
                  file-type
                  '("make-support" "init-comp" "compile-always"
                    "std-heap" "twobit-heap" "r5rs-heap"))))
    (make:project "compiler.date"
      `(rules
        (".lop" ".sch" ,make-compile-and-assemble)
        (".fasl" ".h" ,make-compile-file)  ; wdc
        (".fasl" ".sch" ,make-compile-file))
      `(targets
        ("compiler.date" ,(lambda args #t)))
      `(dependencies
        ("compiler.date" ,compiler-files)
        ("compiler.date" ,comp-util-files)
        ("compiler.date" ,other-util-files)))))

(define compiler-project/fasl (make-compiler-project ".fasl"))

(define (make-compiler . rest)
  (make:pretend (not (null? rest)))
  (make:make compiler-project/fasl "compiler.date"))

(define compiler-project/lop (make-compiler-project ".lop"))

(define (make-petit-compiler . rest)
  (make:pretend (not (null? rest)))
  (make:make compiler-project/lop "compiler.date"))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building all the files in the new generic assembler and
; the Sparc assembler.

(define sparcasm-project
  (let ((common-asm-files
         (append
          (replace-extension ".fasl" (nbuild:common-asm-files))
          (nbuild-files 'common-asm '("link-lop.fasl"))))
        (sparcasm-files
         (replace-extension ".fasl" (nbuild:machine-asm-files))))
    (make:project "sparcasm.date"
      `(rules
        (".fasl" ".h" ,make-compile-file)
        (".fasl" ".sch" ,make-compile-file))
      `(targets
        ("sparcasm.date" ,(lambda args #t)))
      `(dependencies
        ("sparcasm.date" ,common-asm-files)
        ("sparcasm.date" ,sparcasm-files)))))

(define (make-sparcasm . rest)
  (make:pretend (not (null? rest)))
  (make:make sparcasm-project "sparcasm.date"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building all the files in the standard-C assembler.

(define (make-petit-asm-project file-type)
  (let ((common-asm-files
         (replace-extension ".lop" (nbuild:common-asm-files)))
        (petit-asm-files
         (replace-extension ".lop" (nbuild:machine-asm-files))))
    (make:project "petitasm.date"
      `(rules
        (".lop" ".h" ,make-compile-and-assemble)
        (".lop" ".sch" ,make-compile-and-assemble)
        (".fasl" ".sch" ,make-compile-file))
      `(targets
        ("petitasm.date" ,(lambda args #t)))
      `(dependencies
        ("petitasm.date" ,common-asm-files)
        ("petitasm.date" ,petit-asm-files)))))

(define petit-asm-project/fasl (make-petit-asm-project ".fasl"))

(define (make-petitasm . rest)
  (make:pretend (not (null? rest)))
  (make:make petit-asm-project/fasl "petitasm.date"))

(define petit-asm-project/lop (make-petit-asm-project ".lop"))

(define (make-petit-petitasm . rest)
  (make:pretend (not (null? rest)))
  (make:make petit-asm-project/lop "petitasm.date"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building the Larceny compatibility files.

(define (make-compat-project file-type)
  (make:project "compat.date"
    `(rules
      (".lop" ".sch" ,make-compile-and-assemble)
      (".fasl" ".sch" ,make-compile-file))
    `(targets
      ("compat.date" ,(lambda args #t)))
    `(dependencies
      ("compat.date" (,(string-append (nbuild-parameter 'compatibility)
                                      "compat2"
                                      file-type))))))

(define compat-project/fasl (make-compat-project ".fasl"))
(define compat-project/lop (make-compat-project ".lop"))

(define (make-compat . rest)
  (make:pretend (not (null? rest)))
  (make:make compat-project/fasl "compat.date"))

(define (make-petit-compat . rest)
  (make:pretend (not (null? rest)))
  (make:make compat-project/lop "compat.date"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building the Auxiliary libraries.

(define auxlib-files
  '("misc" "list" "vector" "string" "pp" "io" "format" "load" "osdep-unix"))

(define experimental-files
  '("applyhook" "applyhook0" "apropos" "system-stuff"))

(define debugger-files
  '("debug" "countcalls" "trace" "inspect-cont"))

(define (make-auxlib-project file-type)

  (define (auxfile fn)
    (string-append (nbuild-parameter 'auxiliary) 
                   fn 
                   file-type))
  (define (expfile fn)
    (string-append (pathname-append (nbuild-parameter 'root) "Experimental")
                   fn
                   file-type))
  (define (debfile fn)
    (string-append (pathname-append (nbuild-parameter 'root) "Debugger")
                   fn
                   file-type))
                     
  (let ((auxlib-files
         (map auxfile auxlib-files))
        (experimental-files
         (map expfile experimental-files))
        (debugger-files
         (map debfile debugger-files)))
    (make:project "Auxiliary library"
    `(rules
      (".lop" ".sch" ,make-compile-and-assemble)
      (".lop" ".mal" ,make-assemble)
      (".fasl" ".sch" ,make-compile-file)
      (".fasl" ".mal" ,make-assemble-file))
    `(targets
      ("auxlib.date" ,(lambda args #t)))
    `(dependencies
      ("auxlib.date" ,auxlib-files)
      ("auxlib.date" ,experimental-files)
      ("auxlib.date" ,debugger-files)))))

(define auxlib-project/fasl (make-auxlib-project ".fasl"))
(define auxlib-project/lop (make-auxlib-project ".lop"))

(define (make-auxlib . rest)
  (make:pretend (not (null? rest)))
  (make:make auxlib-project/fasl "auxlib.date"))

(define (make-petit-auxlib . rest)
  (make:pretend (not (null? rest)))
  (make:make auxlib-project/lop "auxlib.date"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Project for building the GC test suite.

(define gc-testsuite-project 
  (let ((gc-testsuite-files
         '("dynamic" "gcbench0" "gcbench1" "grow" "lattice" "nbody"
           "nboyer" "nucleic2" "permsort" "sboyer" "dummy")))
    (make:project "GC Testsuite"
      `(rules
        (".fasl" ".sch" ,make-compile-file))
      `(targets
        ("gc-testsuite.date" ,(lambda args #t)))
      `(dependencies
        ("gc-testsuite.date"
         ,(objects "Testsuite/GC/" ".fasl" gc-testsuite-files))))))

(define (make-gc-testsuite . rest)
  (make:pretend (not (null? rest)))
  (make:make gc-testsuite-project "gc-testsuite.date"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Regression tests.

(define regression-test-project
  (let ((regression-test-files 
         '("test" "bool" "char" "ctak" "dynamic-wind" "fact" "fib" "fixnums" 
           "number" "pred" "regression")))
    (make:project "Regression tests"
      `(rules
        (".fasl" ".sch" ,make-compile-file)
        (".fasl" ".mal" ,make-assemble-file))
      `(targets
        ("regression-test.date" ,(lambda args #t)))
      `(dependencies
        ("regression-test.date"
         ,(objects "Testsuite/Lib/" ".fasl" regression-test-files))))))

(define (make-regression-test . rest)
  (make:pretend (not (null? rest)))
  (make:make regression-test-project "regression-test.date"))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Rebuild the entire development system, for all targets.

; Compile to FASL

(define (make-development-environment . rest)
  (apply make-auxlib rest)
  (apply make-compiler rest)
  (apply make-sparcasm rest)
  (apply make-compat rest)
  (compile-file "Lib/makefile.sch"))

; Compile to LOP

(define (make-petit-development-environment . rest)
  (apply make-petit-auxlib rest)
  (apply make-petit-compiler rest)
  (apply make-petit-petitasm rest)
  (apply make-petit-compat rest)
  (compile-and-assemble313 (string-append (nbuild-parameter 'source)
                                          "makefile.sch")))

'  ; Don't remember where this comes from.
(define (make-twobit-application)

  (define (fix files)
    (replace-extension ".lop" files))

  (make-petit-development-environment)
  (build-application 
   "twobit" 
   (append (list 
            (string-append (nbuild-parameter 'compatibility) "compat2.lop")
            (string-append (nbuild-parameter 'auxiliary) "list.lop")
            (string-append (nbuild-parameter 'auxiliary) "pp.lop"))
           (fix (nbuild:twobit-files))
           (fix (nbuild:common-asm-files))
           (fix (nbuild:machine-asm-files))
           (fix (nbuild:heap-dumper-files))
           (fix (nbuild:utility-files)))))

; eof
