; $Id: prefix.sch,v 1.6 2000/02/15 18:17:47 lth Exp $
;
; This file provides an enhanced version of run-benchmark, compatible with
; both the old and new versions of run-benchmark.  Instead of printing
; the results in human-readable form, this version prints them in machine-
; readable form.
;
; Also included is a procedure 'run' that is invoked by the shell scripts
; to run a particular benchmark program.

(define (run benchmark-name iterations . args)

  (define benchmarks
    `((dummy        ,(lambda () (apply dummy-benchmark args)))
      (dynamic      ,(lambda () (apply dynamic-benchmark args)))
      (earley       ,(lambda () (apply earley-benchmark args)))
      (gcbench      ,(lambda () (apply gc-benchmark args)))
      (graphs       ,(lambda () (apply graphs-benchmark args)))
      (lattice      ,(lambda () (apply lattice-benchmark args)))
      (nboyer       ,(lambda () (apply nboyer-benchmark args)))
      (nucleic      ,(lambda () (apply nucleic-benchmark args)))
      (perm8        ,(lambda () (apply perm8-benchmark args)))
      (perm9        ,(lambda () (apply perm9-benchmark args)))
      (sboyer       ,(lambda () (apply sboyer-benchmark args)))
      (softscheme   ,(lambda () (apply softscheme-benchmark args)))
      (textsearch   ,(lambda () (apply textsearch-benchmark args)))
      (twobitlong   ,(lambda () (apply twobit-benchmark 'long args)))
      (twobitshort  ,(lambda () (apply twobit-benchmark 'short args)))
      (vnboyer      ,(lambda () (apply vnboyer-benchmark args)))
      ))

  (let ((probe (assq benchmark-name benchmarks)))
    (if (not probe)
        (begin (display "ERROR The benchmark ")
               (display benchmark-name)
               (display " is not known.  Valid choices are ")
               (display (map car benchmarks))
               (exit))
        (let ((bm (cadr probe)))
          (collect)
          (do ((i 0 (+ i 1)))
              ((= i iterations))
            (bm))))))


; Adapted from Lib/Common/memstats.sch to display more numbers and to 
; display them on a machine-readable form.

; Run-benchmark
;
; (run-benchmark name thunk)
; (run-benchmark name thunk iterations)
; (run-benchmark name iterations thunk ok?)
;
; The 2/3 arg version is compatible with old Larceny releases, and is
; deprecated: a warning is printed if it is used.
;
; The 4 arg version is compatible with benchmarking code that comes 
; with Gambit-C (and superior, because it tests the final value).

(define *nesting* 0)

(define (run-benchmark . args)
  (dynamic-wind
   (lambda () (set! *nesting* (+ *nesting* 1)))
   (lambda ()
     (case (length args)
       ((2 3) (newline)
              ;(display "WARNING: Using old run-benchmark.")
              ;(newline)
              (apply old-run-benchmark args))
       ((4)   (apply new-run-benchmark args))
       (else  (error "Wrong number of arguments to run-benchmark."))))
   (lambda () (set! *nesting* (- *nesting* 1)))))

(define (new-run-benchmark name iterations thunk ok?)

  (define (loop n last-result)
    (if (zero? n)
	last-result
	(loop (- n 1) (thunk))))

  (newline)
  (display "--------------------------------------------------------")
  (newline)
  (display name)
  (newline)
  (let ((result #f))
    (run-with-stats (lambda ()
		      (set! result (loop iterations (unspecified)))))
    (if (not (ok? result))
	(error "Benchmark program returned wrong result: " result))
    (unspecified)))

(define (old-run-benchmark name thunk . rest)
  (let ((n (if (null? rest) 1 (car rest))))
    
    (define (loop n)
      (if (zero? n)
	  #t
	  (begin (thunk)
		 (loop (- n 1)))))

    (newline)
    (display "--------------------------------------------------------")
    (newline)
    (display name)
    (newline)
    (run-with-stats (lambda () (loop n)))))


; This is new

(define (run-with-stats thunk)

  (define (gcinf pred?)
    (let ((dynamic '()))
      (for-each 
       (lambda (area)
         (if (pred? (vector-ref area 1) (vector-ref area 3))
             (set! dynamic (cons (vector-ref area 0) dynamic))))
       (vector->list (cdr (assq 'heap-area-info (system-features)))))
      dynamic))

  (define (indices-of-major-areas)
    (gcinf (lambda (typ inf)
             (or (and (eq? typ 'two-space) (eq? inf 'expandable))
                 (eq? typ 'conservative-mark-sweep)
                 (eq? typ 'two-space-np-old)
                 (eq? typ 'two-space-np-young)
                 (eq? typ 'one-space-dof)))))

  (define (indices-of-minor-areas)
    (gcinf (lambda (typ inf)
             (or (and (eq? typ 'two-space) (eq? inf 'fixed))
                 (eq? typ 'nursery)))))

  (define (gcaccum s1 s2 areas . indices)
    (let ((gcs0 0)
          (gcs1 0))
      (for-each 
       (lambda (a)
         (let ((x0 (vector-ref (vector-ref s1 7) a))
               (x1 (vector-ref (vector-ref s2 7) a)))
           (set! gcs0 (apply + gcs0 
                             (map (lambda (x) (vector-ref x0 x)) indices)))
           (set! gcs1 (apply + gcs1 
                             (map (lambda (x) (vector-ref x1 x)) indices)))))
       areas)
      (- gcs1 gcs0)))

  (define genv.collections 0)
  (define genv.promotions 1)
  (define genv.time-total 2)
  (define genv.time-promotion 10)

  (define (count-minor-gcs s1 s2 minor-areas major-areas)
    (+ (gcaccum s1 s2 minor-areas genv.collections genv.promotions)
       (gcaccum s1 s2 major-areas genv.promotions)))

  (define (count-major-gcs s1 s2 major-areas)
    (gcaccum s1 s2 major-areas genv.collections))

  (define (time-minor-gcs s1 s2 minor-areas major-areas)
    (+ (gcaccum s1 s2 minor-areas genv.time-total)
       (gcaccum s1 s2 major-areas genv.time-promotion)))

  (define (time-major-gcs s1 s2 major-areas)
    (- (gcaccum s1 s2 major-areas genv.time-total)
       (gcaccum s1 s2 major-areas genv.time-promotion)))

  (define (print-stats s1 s2)
    (let ((major (indices-of-major-areas))
          (minor (indices-of-minor-areas)))
      (let* ((allocated (- (vector-ref s2 0) (vector-ref s1 0)))
             (reclaimed (- (vector-ref s2 1) (vector-ref s1 1)))
             (copied (- (vector-ref s2 2) (vector-ref s1 2)))
             (elapsed (- (vector-ref s2 21) (vector-ref s1 21)))
             (user (- (vector-ref s2 23) (vector-ref s1 23)))
             (system (- (vector-ref s2 22) (vector-ref s1 22)))
             (gctime (- (vector-ref s2 3) (vector-ref s1 3)))
             (mingctime (time-minor-gcs s1 s2 minor major))
             (majgctime (time-major-gcs s1 s2 major))
             (majgcs (count-major-gcs s1 s2 major))
             (mingcs (count-minor-gcs s1 s2 minor major))
             (gcs    (+ majgcs mingcs)))
        (newline)                       ; Force start of line
        (display "STATS-OUTPUT ")
        (write `((bytes-allocated . ,(* 4 allocated))
                 (bytes-reclaimed . ,(* 4 reclaimed))
                 (bytes-marked . ,(* 4 copied))
                 (elapsed-time . ,elapsed)
                 (cpu-time-user . ,user)
                 (cpu-time-system . ,system)
                 (elapsed-time-gc . ,gctime)
                 (elapsed-time-major-gc . ,majgctime)
                 (elapsed-time-minor-gc . ,mingctime)
                 (count-gc . ,gcs)
                 (count-major-gc . ,majgcs)
                 (count-minor-gc . ,mingcs)
                 (bytes-max-heap-used . ,(* 4 (vector-ref s2 27)))
                 (raw-data-before . ,s1)
                 (raw-data-after . ,s2)))
        (newline))))
  
  (let* ((s1 (memstats))
	 (r  (thunk))
	 (s2 (memstats)))
    (if (= *nesting* 1)
        (print-stats s1 s2))
    r))

(define error 
  (let ((error error))
    (lambda args
      (newline)                         ; For the benefit of parsing output
      (display "ERROR ")
      (apply error args))))

; eof

